from pathlib import Path
from beartype.typing import Iterator, Tuple
from camera_geometry import FrameSet

from camera_geometry.camera_models import optimal_undistorted
import torch


import numpy as np
from splat_trainer.camera_table.camera_table import CameraRigTable, CameraTable
from splat_trainer.dataset.dataset import CameraView, Dataset
from splat_trainer.util.misc import split_stride


from .loading import  PreloadedImages, preload_images
from splat_trainer.util.pointcloud import PointCloud
from .visibility import visibility



class ScanDataset(Dataset):
  def __init__(self, scan_file:str,                
        image_scale:float=1.0,
        val_stride:int=10,
        depth_range:Tuple[float, float] = (0.1, 100.0)):

    self.scan_file = scan_file
    self.image_scale = image_scale

    scan = FrameSet.load_file(Path(scan_file))
    self.depth_range = depth_range

    cameras = {k: optimal_undistorted(camera, alpha=0).scale_image(image_scale)
      for k, camera in scan.cameras.items()}

    print("Undistorted cameras:")
    for k, camera in cameras.items():
        print(k, camera)

    print("Loading images...")
    self.all_cameras = preload_images(scan, cameras)
    self.scan = scan.copy(cameras=cameras)

    # Evenly distribute validation images
    self.train_cameras, self.val_cameras = split_stride(self.all_cameras, val_stride)
    
  def __repr__(self) -> str:
    return f"ScanDataset({self.scan_file}, image_scale={self.image_scale} cloud={find_cloud(self.scan)})"

  def train(self, shuffle=True) -> Iterator[CameraView]:
    images = PreloadedImages(self.train_cameras, shuffle=shuffle)
    return images
    # images = PreloadedImages(self.train_cameras)
    # return torch.utils.data.DataLoader(images, batch_size=1, shuffle=shuffle, pin_memory=True, num_workers=2)
    
  def val(self) -> Iterator[CameraView]:
    images = PreloadedImages(self.val_cameras)
    return images


  def camera_table(self) -> CameraRigTable:
    camera_t_rig = np.array(
       [camera.camera_t_parent for camera in self.scan.cameras.values()])
    
    world_t_rig = torch.from_numpy(np.array(self.scan.rig_poses)).to(torch.float32)
    projections = np.array([camera.intrinsic for camera in self.scan.cameras.values()])

    
    return CameraRigTable(
      rig_t_world=torch.linalg.inv(world_t_rig),
      camera_t_rig=torch.from_numpy(camera_t_rig).to(torch.float32),
      projection=torch.from_numpy(projections).to(torch.float32))
  
  def camera_shape(self) -> torch.Size:
    return torch.Size([self.scan.num_frames, len(self.scan.cameras)])
  

  def pointcloud(self) -> PointCloud:
    pcd_filename = find_cloud(self.scan)    
    pcd = PointCloud.load(pcd_filename)

    counts = visibility(self.scan.expand_cameras(), pcd.points)
    print(f"Visible {(counts > 0).sum()} of {len(counts)} points")
    # pcd = pcd.select_by_index(np.flatnonzero(vis > 0))
    
    return pcd[counts > 0]


def find_cloud(scan:FrameSet) -> Tuple[np.ndarray, np.ndarray]:
  assert 'sparse' in scan.models, "No sparse model found in scene.json"
  return Path(scan.find_file(scan.models.sparse.filename))




from torch import nn
import torch
import torch.nn.functional as F

from taichi_splatting.torch_ops.transforms import split_rt

from splat_trainer.util.transforms import join_rt
import roma



class RigPoseTable(nn.Module):
  def __init__(self, rig_t_world:torch.Tensor, camera_t_rig:torch.Tensor):
    super().__init__()

    self.camera_t_rig = PoseTable(camera_t_rig)
    self.rig_t_world = PoseTable(rig_t_world)

  @property
  def num_cameras(self):
    return self.camera_t_rig.t.shape[0]
  
  @property 
  def num_frames(self):
    return self.rig_t_world.t.shape[0]
  
  @property
  def rig_t_camera(self):
    return torch.linalg.inv(self.camera_t_rig.matrix)
  
  @property
  def world_t_rig(self):
    return torch.linalg.inv(self.rig_t_world.matrix)

  def normalize(self):
    self.camera_t_rig.normalize()
    self.rig_t_world.normalize()

  
  def forward(self, image_index:torch.Tensor):
    assert image_index.dim() == 2 and image_index.shape[1] == 2, \
      f"Expected (rig_index, camera_index) N, 2 tensor, got: {image_index.shape}"

    rig_index, camera_index = image_index.unbind(-1)

    camera_t_rig = self.camera_t_rig(camera_index)
    rig_t_world = self.rig_t_world(rig_index)

    return camera_t_rig @ rig_t_world





class PoseTable(nn.Module):
  def __init__(self, m:torch.Tensor):
    super().__init__()

    assert m.shape[-2:] == (4, 4), f"Expected (..., 4, 4) tensor, got: {m.shape}"


    R, t = split_rt(m)
    q = roma.rotmat_to_unitquat(R)

    self.t = nn.Parameter(t.to(torch.float32))
    self.q = nn.Parameter(q.to(torch.float32))


  def forward(self, indices):
    assert (indices < self.q.shape[0]).all(), f"Index out of bounds: {indices} >= {self.q.shape[0]}"

    q, t = F.normalize(self.q[indices], dim=-1), self.t[indices]
    return join_rt(roma.unitquat_to_rotmat(q), t)
  
  def normalize(self):
    self.q.data = F.normalize(self.q.data, dim=-1)
    return self

  @property
  def shape(self):
    return self.t.shape[:-1]


  

if __name__ == '__main__':
  torch.set_printoptions(precision=6, sci_mode=False)
  torch.manual_seed(0)

  for i in range(10):
    q = F.normalize(torch.randn(10, 4))
    m = roma.unitquat_to_rotmat(q)

    assert roma.rotmat_to_unitquat(m).allclose(q), f"{roma.rotmat_to_unitquat(m)} != {q}"





from abc import abstractmethod, abstractproperty

import numpy as np
import torch
from torch import nn


from splat_trainer.camera_table.pose_table import PoseTable, RigPoseTable


class CameraTable(nn.Module):

  @abstractmethod
  def forward(self, image_idx:torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    raise NotImplementedError
  
  @abstractproperty
  def shape(self) -> torch.Size:
    raise NotImplementedError
  
  @abstractproperty
  def all_cameras(self):
    raise NotImplementedError

  
  @abstractproperty
  def camera_centers(self) -> torch.Tensor:
    raise NotImplementedError
        
  

def camera_extents(cameras:CameraTable):
    cam_centers = cameras.camera_centers.reshape(-1, 3)
    avg_cam_center = torch.mean(cam_centers, dim=0, keepdim=True)

    distances = torch.norm(cam_centers - avg_cam_center, dim=0, keepdim=True)
    diagonal = torch.max(distances)

    return avg_cam_center.reshape(3), (diagonal * 1.1).item()


class CameraRigTable(CameraTable):
  def __init__(self, rig_t_world:torch.Tensor,   # (N, 4, 4) - poses for the whole camera rig
                     camera_t_rig:torch.Tensor,  # (C, 4, 4) - camera poses inside the rig
                     projection:torch.Tensor   # (C, 3, 3) - camera intrinsics for each camera in rig
                    ):
    super().__init__()

    self.camera_projection = torch.nn.Parameter(
        projection.to(torch.float32))
    self.camera_poses = RigPoseTable(
      rig_t_world=rig_t_world, camera_t_rig=camera_t_rig)


  def forward(self, image_idx:torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:     
    return self.camera_poses(image_idx.unsqueeze(0)), self.camera_projection[image_idx[1]]

  @property
  def shape(self) -> torch.Size:
    return torch.Size([self.camera_poses.num_frames, self.camera_poses.num_cameras])
  
  @property
  def camera_centers(self) -> torch.Tensor:

    world_t_camera = torch.linalg.inv(self.camera_poses(self.all_cameras))
    return world_t_camera[..., :3, 3]
  
  @property
  def device(self):
    return self.camera_projection.device
  
  @property
  def all_cameras(self):
    dims = [torch.arange(0, n, device=self.device) for n in self.shape]
    return torch.stack(torch.meshgrid(*dims, indexing='ij'), dim=-1).view(-1, 2)



class MultiCameraTable(CameraTable):
  """
  A table of camera poses and intrinsics - cameras can have different intrinsics which are stored in the projection table.
  """
  def __init__(self, 
               camera_t_world:torch.Tensor, # (N, 4, 4)
               camera_idx:torch.Tensor,     # (N,) - index into projection table (0, P-1)
               projection:torch.Tensor     # (P, 4, 4),
              ):
    super().__init__()

    self.camera_projection = torch.nn.Parameter(
      projection.to(torch.float32))
    self.register_buffer("camera_idx", camera_idx)
    self.camera_t_world = PoseTable(camera_t_world)


  def forward(self, image_idx:torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
      cam_idx = self.camera_idx[image_idx]
      return self.camera_t_world(image_idx), self.camera_projection[cam_idx]

  @property
  def shape(self) -> torch.Size:
    return self.camera_t_world.shape
  
  @property
  def camera_centers(self) -> torch.Tensor:
    world_t_camera = torch.linalg.inv(self.camera_t_world(self.all_cameras))
    return world_t_camera[..., :3, 3]
  
  @property
  def device(self):
    return self.camera_projection.device

  @property
  def all_cameras(self):
    return torch.arange(0, self.shape[0], device=self.camera_projection.device)




from pathlib import Path
import tempfile
import plyfile

import numpy as np
import torch
import torch.nn.functional as F

from taichi_splatting import Gaussians3D


def to_plydata(gaussians:Gaussians3D, with_sh:bool=False) -> plyfile.PlyData:
  gaussians = gaussians.to('cpu')


  dtype = [
    ('x', 'f4'), ('y', 'f4'), ('z', 'f4'),
    ('opacity', 'f4'),
    ('scale_0', 'f4'), ('scale_1', 'f4'), ('scale_2', 'f4'),
    ('rot_0', 'f4'), ('rot_1', 'f4'), ('rot_2', 'f4'), ('rot_3', 'f4'),
   ] 
  
  if with_sh is True:
    assert gaussians.feature.ndim == 3, f"Expected ndim=3 sh_feature tensor, got {gaussians.feature.shape}"
    num_sh = gaussians.feature.shape[2] *  gaussians.feature.shape[1]


    dtype += [('f_dc_0', 'f4'), ('f_dc_1', 'f4'), ('f_dc_2', 'f4')] + [
      ('f_rest_' + str(i), 'f4') for i in range(num_sh - 3)]
  else:
    assert gaussians.feature.ndim == 2, f"Expected ndim=2 feature tensor, got {gaussians.feature.shape}"

    num_features = gaussians.feature.shape[1] 
    dtype += [('f_' + str(i), 'f4') for i in range(num_features)]

  vertex = np.zeros(gaussians.batch_size[0], dtype=dtype )

  for i, name in enumerate(['x', 'y', 'z']):
    vertex[name] = gaussians.position[:, i].numpy()

  for i in range(3):
    vertex['scale_' + str(i)] = gaussians.log_scaling[:, i].numpy()

  rotation = torch.roll(F.normalize(gaussians.rotation), 1, dims=(1,))
  for i in range(4):
    vertex['rot_' + str(i)] = rotation[:, i].numpy()

  vertex['opacity'] = gaussians.alpha_logit[:, 0].numpy()

  if with_sh is True:
    sh_dc, sh_rest = gaussians.feature[:, :, 0], gaussians.feature[:, :, 1:]

    sh_dc = sh_dc.view(-1, 3)
    sh_rest = sh_rest.reshape(sh_rest.shape[0], sh_rest.shape[1] * sh_rest.shape[2])

    for i in range(3):
      vertex['f_dc_' + str(i)] = sh_dc[:, i].numpy()

    for i in range(sh_rest.shape[1]):
      vertex['f_rest_' + str(i)] = sh_rest[:, i].numpy()
  else:
    for i in range(num_features):
      vertex['f_' + str(i)] = gaussians.feature[:, i].numpy()

  el = plyfile.PlyElement.describe(vertex, 'vertex')
  return plyfile.PlyData([el])


def from_plydata(plydata:plyfile.PlyData, with_sh:bool=False) -> Gaussians3D:
  
  vertex = plydata['vertex']

  def get_keys(ks):
    values = [torch.from_numpy(vertex[k].copy()) for k in ks]
    return torch.stack(values, dim=-1)

  
  positions = torch.stack(
    [ torch.from_numpy(vertex[i].copy()) for i in ['x', 'y', 'z']], dim=-1)

  attrs = sorted(plydata['vertex'].data.dtype.names)
  log_scaling = get_keys([f'scale_{k}' for k in range(3)])
  
  if with_sh is True:
    sh_attrs = [k for k in attrs if k.startswith('f_rest_') or k.startswith('f_dc_')]
    
    n_sh = len(sh_attrs) // 3
    deg = int(np.sqrt(n_sh))

    assert deg * deg == n_sh, f"SH feature count must be square ({deg} * {deg} != {n_sh}), got {len(sh_attrs)}"

    sh_dc = get_keys([f'f_dc_{k}' for k in range(3)]).view(positions.shape[0], 3, 1)
    sh_rest = get_keys([f'f_rest_{k}' for k in range(3 * (deg * deg - 1))])
    sh_rest = sh_rest.view(positions.shape[0], 3, n_sh - 1)

    features = torch.cat([sh_dc, sh_rest], dim=2)  
  else:
    feature_attrs = [k for k in attrs if k.startswith('f_')]
    features = get_keys([f'f_{k}' for k in range(len(feature_attrs))])

  rotation = get_keys([f'rot_{k}' for k in range(4)])
  # convert from wxyz to xyzw quaternion and normalize
  rotation = torch.roll(F.normalize(rotation), -1, dims=(1,))
  
  alpha_logit = get_keys(['opacity'])

  
  return Gaussians3D(
    position = positions, 
    rotation = rotation,
    alpha_logit = alpha_logit,
    log_scaling = log_scaling,
    feature = features,

    batch_size = (positions.shape[0],)
  )

def write_gaussians(filename:Path | str, gaussians:Gaussians3D, with_sh:bool = False):
  filename = Path(filename)

  plydata = to_plydata(gaussians.detach(), with_sh=with_sh)
  plydata.write(str(filename))



def read_gaussians(filename:Path | str, with_sh:bool = False) -> Gaussians3D:
  filename = Path(filename) 

  plydata = plyfile.PlyData.read(str(filename))
  return from_plydata(plydata, with_sh=with_sh)


  

def random_gaussians(n:int, sh_degree:int):
  points = torch.randn(n, 3)

  return Gaussians3D( 
    position = points,
    rotation = F.normalize(torch.randn(n, 4), dim=1),
    alpha_logit = torch.randn(n, 1),
    log_scaling = torch.randn(n, 3) * 4,
    feature = torch.randn(n, 3, (sh_degree + 1)**2),

    batch_size = (n,)
  )

def test_read_write():
  temp_path = Path(tempfile.mkdtemp())

  print("Testing write/read")
  for i in range(10):
    g = random_gaussians((i + 1) * 1000, 3)
    write_gaussians(temp_path / f'gaussians_{i}.ply', g)
    g2 = read_gaussians(temp_path / f'gaussians_{i}.ply')

    assert torch.allclose(g.position, g2.position)
    assert torch.allclose(g.rotation, g2.rotation)
    assert torch.allclose(g.alpha(), g2.alpha())
    assert torch.allclose(g.log_scaling, g2.log_scaling)
    assert torch.allclose(g.feature, g2.feature)





if __name__ == '__main__':
  test_read_write()





import numpy as np
from scipy.spatial import cKDTree
import torch
import torch.nn.functional as F

from splat_trainer.util.misc import inverse_sigmoid
from splat_trainer.util.pointcloud import PointCloud


from taichi_splatting import Gaussians3D
  

def from_pointcloud(pcd:PointCloud, 
                    initial_scale:float = 0.5,
                    initial_alpha:float = 0.5,
                    num_neighbors:int = 3) -> Gaussians3D:
  scales = estimate_scale(pcd, num_neighbors=num_neighbors) * initial_scale

  return Gaussians3D(
    position=pcd.points,
    log_scaling=torch.log(scales).unsqueeze(1).expand(-1, 3),
    rotation=F.normalize(torch.randn(pcd.points.shape[0], 4), dim=1),
    alpha_logit=torch.full( (pcd.points.shape[0], 1), 
                            fill_value=inverse_sigmoid(initial_alpha)),
    feature=pcd.colors,
    batch_size=(pcd.points.shape[0],)
  )

  
def estimate_scale(pointcloud : PointCloud, num_neighbors:int = 3):
  points = pointcloud.points.cpu().numpy()
  tree = cKDTree(points)

  dist, idx = tree.query(points, k=num_neighbors + 1, workers=-1)
  
  distance = np.mean(dist[:, 1:], axis=1)
  return torch.from_numpy(distance).to(torch.float32)



import argparse
from time import time
import math
import numpy as np
import tinycudann as tcnn
import torch
import torch.nn.functional as F
from tqdm import tqdm





def main():
  parser = argparse.ArgumentParser("Benchmark tcnn")
  parser.add_argument("--n", type=int, default=1000000, help="Number of points")
  parser.add_argument("--hidden", type=int, default=32, help="Number of hidden neurons")
  parser.add_argument("--layers", type=int, default=2, help="Number of hidden layers")
  parser.add_argument("--num_features", type=int, default=8, help="Number of features")
  

  args = parser.parse_args()

  model = tcnn.NetworkWithInputEncoding(
    args.num_features + 3, 3,
    encoding_config=dict(
      otype = "composite",
      nested = [
        dict(otype = "SphericalHarmonics", 
            degree = 4, 
            n_dims_to_encode = 3
        ),
        
        dict(otype = "Identity",
            n_dims_to_encode = args.num_features)
      ]
    ), 
    
    network_config = dict(
      otype = "FullyFusedMLP",
      activation = "ReLU",
      output_activation = "None",
      n_neurons = args.hidden,
      n_hidden_layers = args.layers,
    )
  )


  
  device = "cuda:0"

  dir = F.normalize(torch.randn( (args.n, 3) ), dim=1)
  features = torch.cat([dir, torch.randn(args.n, args.num_features)], dim=1).to(device)

  features.requires_grad_(True)

  iters = 100
  start = time()
  
  pbar = tqdm(desc = f"n={args.n}", total=iters)
  for i in range(iters):
    out = model.forward(features)

    err = out.sum()
    err.backward()

    pbar.update(1)

  torch.cuda.synchronize()
  end = time()

  print(f"N: {args.n} Rate: {iters / (end - start)}")


  

if __name__ == "__main__":
  main()


from .scene import GaussianSceneConfig, GaussianScene
from .sh_scene import SHScene, Gaussians3D, SHConfig
from .tcnn_scene import TCNNScene, TCNNConfig


__all__ = [
  "GaussianScene",
  "GaussianSceneConfig",

  "SHScene",
  "SHConfig",
  
  "TCNNScene",
  "TCNNConfig",

  "Gaussians3D"
]



from abc import ABCMeta, abstractmethod
from collections import namedtuple
from beartype.typing import Iterator

import torch

from splat_trainer.camera_table.camera_table import CameraTable
from splat_trainer.util.pointcloud import PointCloud


CameraView = namedtuple('CameraView', 
  ('filename', 'image', 'index'))

class Dataset(metaclass=ABCMeta):
  
  @abstractmethod
  def train(self, shuffle=True) -> Iterator[CameraView]:
    raise NotImplementedError
    
  @abstractmethod
  def val(self) -> Iterator[CameraView]:
    raise NotImplementedError

  @abstractmethod
  def camera_table(self) -> CameraTable:
    raise NotImplementedError


  @abstractmethod
  def pointcloud(self) -> PointCloud:
    raise NotImplementedError



from beartype.typing import Tuple
import torch
import torch.nn.functional as F




def split_rt(
    transform: torch.Tensor,  # (batch_size, 4, 4)
) -> Tuple[torch.Tensor, torch.Tensor]:
    R = transform[..., :3, :3]
    t = transform[..., :3, 3]
    return R.contiguous(), t.contiguous()

def join_rt(r, t):
  assert r.shape[-2:] == (3, 3), f"Expected (..., 3, 3) tensor, got: {r.shape}"
  assert t.shape[-1] == 3, f"Expected (..., 3) tensor, got: {t.shape}"

  prefix = t.shape[:-1]
  assert prefix == t.shape[:-1], f"Expected same prefix shape, got: {r.shape} {t.shape}"

  T = torch.eye(4, device=r.device, dtype=r.dtype).view((1, ) * (len(prefix)) + (4, 4)).expand(prefix + (4, 4)).contiguous()

  T[..., 0:3, 0:3] = r
  T[..., 0:3, 3] = t
  return T
  

def make_homog(points):
  shape = list(points.shape)
  shape[-1] = 1
  return torch.concatenate([points, torch.ones(shape, dtype=points.dtype, device=points.device)], axis=-1)

def transform44(transform, points):

  points = points.reshape([-1, 4, 1])
  transformed = transform.reshape([1, 4, 4]) @ points
  return transformed[..., 0].reshape(-1, 4)



def transform33(transform, points):

  points = points.reshape([-1, 3, 1])
  transformed = transform.reshape([1, 3, 3]) @ points
  return transformed[..., 0].reshape(-1, 3)





from abc import ABCMeta, abstractmethod
from taichi_splatting import Rendering
import torch

from splat_trainer.scene import GaussianScene

class ControllerConfig(metaclass=ABCMeta):

  @abstractmethod
  def make_controller(self, scene:GaussianScene, 
                      densify_interval:int, total_steps:int) -> 'Controller':
    raise NotImplementedError
  

class Controller(metaclass=ABCMeta):
  @abstractmethod
  def densify_and_prune(self, step:int, total_steps:int):
    raise NotImplementedError

  @abstractmethod
  def add_rendering(self, rendering:Rendering) -> tuple[torch.Tensor, torch.Tensor]: 
    raise NotImplementedError
  
  @abstractmethod
  def log_histograms(self, step:int):
    raise NotImplementedError





from .dataset import Dataset, CameraView
from .colmap.dataset import COLMAPDataset

def ScanDataset(*args, **kwargs):
    from .scan.dataset import ScanDataset
    return ScanDataset(*args, **kwargs)


__all__ = [
    "CameraView",
    "Dataset",
    
    "ScanDataset",  
    "COLMAPDataset"
    ]



from dataclasses import replace
from functools import partial
import math
from beartype.typing import Optional
from beartype import beartype

import torch
import torch.nn.functional as F
from taichi_splatting import Gaussians3D
import roma
    

def point_basis(points:Gaussians3D):
  scale = torch.exp(points.log_scaling)

  r = F.normalize(points.rotation, dim=1)
  m = roma.unitquat_to_rotmat(r)


  return m.transpose(1, 2) * scale.unsqueeze(-1)


def split_by_samples(points: Gaussians3D, samples: torch.Tensor) -> Gaussians3D:
  num_points, n, _ = samples.shape

  basis = point_basis(points)
  point_samples = (samples.view(-1, 3).unsqueeze(1) @ basis.repeat_interleave(repeats=n, dim=0)).squeeze(1)

  gaussians = points.apply(
    partial(torch.repeat_interleave, repeats=n, dim=0), 
    batch_size=[num_points * n])
  
  return replace(gaussians,
    position = gaussians.position + point_samples,
    batch_size=(num_points * n, ))
   

@beartype
def split_gaussians(points: Gaussians3D, n:int=2, scaling:Optional[float]=None) -> Gaussians3D:
  """
  Toy implementation of the splitting operation used in gaussian-splatting,
  returns a scaled, randomly sampled splitting of the gaussians.

  Args:
      points: The Gaussians3D parameters to be split
      n: number of gaussians to split into
      scale: scale of the new gaussians relative to the original

  Returns:
      Gaussians2D: the split gaussians 
  """

  samples = torch.randn((points.batch_size[0], n, 3), device=points.position.device) 

  if scaling is None:
    scaling = 1 / math.sqrt(n)

  points = replace(points, 
      log_scaling = points.log_scaling + math.log(scaling),
      batch_size = points.batch_size)

  return split_by_samples(points, samples)


def split_gaussians_uniform(points: Gaussians3D, n:int=2, scaling:Optional[float]=None, noise=0.1) -> Gaussians3D:
  """ Split along most significant axis """
  axis = F.one_hot(torch.argmax(points.log_scaling, dim=1), num_classes=3)
  values = torch.linspace(-1, 1, n, device=points.position.device)

  samples = values.view(1, -1, 1) * axis.view(-1, 1, 3)
  samples += torch.randn_like(samples) * noise

  if scaling is None:
    scaling = 1 / math.sqrt(n)

  points = replace(points, 
      log_scaling = points.log_scaling + math.log(scaling) * axis,
      batch_size = points.batch_size)

  return split_by_samples(points, samples)



import copy
from dataclasses import  dataclass, replace
from functools import partial
from pathlib import Path
from beartype import beartype
from omegaconf import DictConfig, OmegaConf

import torch
import torch.nn.functional as F
import tinycudann as tcnn
from tqdm import tqdm

from splat_trainer.camera_table.camera_table import CameraTable, camera_extents
from splat_trainer.logger.logger import Logger
from splat_trainer.scene.io import write_gaussians
from splat_trainer.scene.scene import GaussianSceneConfig, GaussianScene
from splat_trainer.gaussians.split import split_gaussians_uniform


from taichi_splatting.misc.parameter_class import ParameterClass
from taichi_splatting import Gaussians3D, RasterConfig, Rendering

from taichi_splatting.renderer import gaussians_in_view, project_to_image, render_projected
from taichi_splatting.perspective import CameraParams

from splat_trainer.util.pointcloud import PointCloud


@dataclass(kw_only=True, frozen=True)
class TCNNConfig(GaussianSceneConfig):  
  learning_rates : DictConfig
  
  lr_image_feature:float= 0.001
  lr_nn:float = 0.0001

  image_features:int       = 8
  point_features:int       = 8

  hidden:int             = 32
  layers:int             = 2


  def color_model(self):
    num_features = self.image_features + self.point_features

    return tcnn.NetworkWithInputEncoding(
      num_features + 3, 3,
      encoding_config=dict(
        otype = "composite",
        nested = [
          dict(otype = "SphericalHarmonics", 
              degree = 4, 
              n_dims_to_encode = 3
          ),
          
          dict(otype = "Identity",
              n_dims_to_encode = num_features)
        ]
      ), 
      
      network_config = dict(
        otype = "FullyFusedMLP",
        activation = "ReLU",
        output_activation = "None",
        n_neurons = self.hidden,
        n_hidden_layers = self.layers,
      )
    )


  def with_scene_extent(self, scene_extent:float) -> 'TCNNConfig':
    learning_rates = copy.copy(self.learning_rates) 
    learning_rates.position *= scene_extent
    return replace(self, learning_rates=learning_rates)
  


  def from_color_gaussians(self, gaussians:Gaussians3D, camera_table:CameraTable, device:torch.device):
    color_model = self.color_model().to(device)
    color_targets = gaussians.feature.to(device)

    features = torch.zeros(gaussians.batch_size[0], self.point_features)
    gaussians = gaussians.replace(feature=features).to(device)

    self.pretrain_colors(gaussians.feature, color_model, color_targets, iters=1000)
    
    centre, extent = camera_extents(camera_table)
    config = self.with_scene_extent(extent)

    return TCNNScene(gaussians, color_model, camera_table, device, config)


  def pretrain_colors(self, features, color_model, target_colors, iters):
    with torch.enable_grad():
      n = target_colors.shape[0]
      device = features.device

      features = torch.nn.Parameter(features, requires_grad=True)
      param_groups = [
        dict(params=color_model.parameters(), lr=self.lr_nn, name="color_model"),
        dict(params=features, lr=self.learning_rates["feature"], name="point_features")
      ]

      opt = torch.optim.Adam(param_groups)

      pbar = tqdm(total=iters, desc="initializing colors")
      for i in range(iters):
        opt.zero_grad()
        dirs = F.normalize(torch.randn(n, 3, device=device), dim=1)
        cam_feature = torch.zeros(n, self.image_features, device=device)

        feature = torch.cat([dirs, features, cam_feature], dim=1)
        colors = color_model(feature).to(torch.float32).sigmoid()

        loss = F.mse_loss(colors, target_colors)

        loss.backward()
        opt.step()
        pbar.update(1)
        pbar.set_postfix(loss=loss.item())


@beartype
class TCNNScene(GaussianScene):
  def __init__(self, points: Gaussians3D, 
               color_model:torch.nn.Module,
               camera_table:CameraTable,

               device:torch.device,            
               config: TCNNConfig):
    self.config = config
    self.device = device

    self.color_model = color_model

    self.raster_config = RasterConfig()
    self.learning_rates = OmegaConf.to_container(config.learning_rates)

    make_optimizer = partial(torch.optim.SparseAdam, betas=(0.7, 0.999))
    self.points = ParameterClass.create(points.to_tensordict(), 
          learning_rates = self.learning_rates, optimizer=make_optimizer)   
    
    self.camera_table = camera_table
    
    image_features = torch.zeros(*camera_table.shape,  config.image_features, dtype=torch.float32, device=device)
    self.image_features = torch.nn.Parameter(image_features, requires_grad=True)

    self.color_opt = self.make_color_optimizer()

  def make_color_optimizer(self):
    config = self.config

    param_groups = [
      dict(params=self.color_model.parameters(), lr=config.lr_nn, name="color_model"),
      dict(params=self.image_features, lr=config.lr_image_feature, name="image_features")
    ]

    return torch.optim.Adam(param_groups)

  
  
  @property
  def num_points(self):
    return self.points.position.shape[0]

  @beartype
  def update_learning_rate(self, lr_scale:float):
    # scaled_lr = {k: v * lr_scale for k, v in self.learning_rates.items()}
    # self.points.set_learning_rate(**scaled_lr)
    self.points.set_learning_rate(position = self.learning_rates ['position'] * lr_scale)


  def __repr__(self):
    return f"GaussianScene({self.points.position.shape[0]} points)"

  def step(self, visible:torch.Tensor):
    
    self.points.step_sparse(visible)
    self.color_opt.step()

    self.points.rotation = torch.nn.Parameter(
      F.normalize(self.points.rotation.detach(), dim=1), requires_grad=True)

  def zero_grad(self):
    self.points.zero_grad()


  def split_and_prune(self, keep_mask, split_idx):
    splits = split_gaussians_uniform(self.gaussians[split_idx], n=2)
    self.points = self.points[keep_mask].append_tensors(splits.to_tensordict())
    return self

  @property
  def gaussians(self):
      return Gaussians3D.from_tensordict(self.points.tensors)
      
  def evaluate_colors(self, indexes, image_idx, camera_position):
    cam_feature = self.image_features[image_idx.unbind(0)]  
    cam_feature = cam_feature.unsqueeze(0).expand(indexes.shape[0], -1)

    dir = F.normalize(self.points.position[indexes].detach() - camera_position)

    feature = torch.cat([dir, self.points.feature[indexes], cam_feature], dim=1)
    return self.color_model(feature).to(torch.float32).sigmoid()
  

  def random_camera(self):
      image_idx = [torch.randint(0, j, device=self.device) for j in self.image_features.shape]
      return torch.stack(image_idx)


  def write_to(self, output_dir:Path, base_filename:str):
    output_dir.mkdir(parents=True, exist_ok=True)
    write_gaussians(output_dir / f'{base_filename}.ply', self.gaussians.detach(), with_sh=False)

  def log(self, logger:Logger, step:int):
    image_idx = torch.zeros(len(self.camera_table.shape), device=self.device, dtype=torch.long)
    camera_position = self.camera_table.camera_centers[image_idx.unbind(0)]

    colors = self.evaluate_colors(torch.arange(self.num_points, device=self.device), image_idx, camera_position)

    point_cloud = PointCloud(self.gaussians.position.detach(), colors)
    logger.log_cloud("point_cloud", point_cloud, step=step)
    


  def render(self, camera_params:CameraParams, image_idx:torch.Tensor, **options) -> Rendering:
    
    config = self.raster_config
    indexes = gaussians_in_view(self.points.position, camera_params, config.tile_size, config.margin_tiles)

    features = self.evaluate_colors(indexes, image_idx, camera_params.camera_position)
    gaussians2d, depthvars = project_to_image(self.gaussians, indexes, camera_params)


    return render_projected(indexes, gaussians2d, features, depthvars, 
            camera_params, config, **options)







from .logger import Logger
from .tensorboard import TensorboardLogger

def WandbLogger(*args, **kwargs):
    from .wandb import WandbLogger
    return WandbLogger(*args, **kwargs)


__all__ = ["WandbLogger", "TensorboardLogger", "Logger"]



import copy
from dataclasses import  asdict, dataclass, replace
from pathlib import Path
from beartype import beartype
from omegaconf import DictConfig, OmegaConf

import torch
import torch.nn.functional as F

from splat_trainer.camera_table.camera_table import CameraTable, camera_extents
from splat_trainer.logger.logger import Logger
from splat_trainer.scene.io import write_gaussians
from splat_trainer.scene.scene import GaussianSceneConfig, GaussianScene
from splat_trainer.gaussians.split import  split_gaussians_uniform
from splat_trainer.util.misc import  rgb_to_sh, sh_to_rgb

from taichi_splatting.misc.parameter_class import ParameterClass
from taichi_splatting import Gaussians3D, RasterConfig, render_gaussians, Rendering
from taichi_splatting.perspective import CameraParams


from splat_trainer.util.pointcloud import PointCloud


@dataclass(kw_only=True, frozen=True)
class SHConfig(GaussianSceneConfig):  
  learning_rates : DictConfig
  sh_ratio:float      = 20.0
  sh_degree:int       = 2


  def with_scene_extent(self, scene_extent:float) -> 'SHConfig':
    learning_rates = copy.copy(self.learning_rates) 
    learning_rates.position *= scene_extent
    return replace(self, learning_rates=learning_rates)
    

  def from_color_gaussians(self, gaussians:Gaussians3D, camera_table:CameraTable, device:torch.device):
    sh_feature = torch.zeros(gaussians.batch_size[0], 3, (self.sh_degree + 1)**2)
    sh_feature[:, :, 0] = rgb_to_sh(gaussians.feature)

    centre, extent = camera_extents(camera_table)
    config = self.with_scene_extent(extent)
    
    return SHScene(gaussians.replace(feature=sh_feature), camera_table, device, config)


class SHScene(GaussianScene):
  def __init__(self, points: Gaussians3D, camera_table:CameraTable, device:torch.device, config: SHConfig):
    self.config = config
    self.camera_table = camera_table

    self.raster_config = RasterConfig()
    self.learning_rates = OmegaConf.to_container(config.learning_rates)

    self.points = ParameterClass.create(points.to_tensordict().to(device), 
          learning_rates = self.learning_rates)   
        
  def with_scene_extent(self, scene_extent:float) -> 'SHScene':
    learning_rates = copy.copy(self.learning_rates) 
    learning_rates.position *= scene_extent
    return replace(self, learning_rates=learning_rates)
  
  @property 
  def device(self):
    return self.points.position.device
  
  @property
  def num_points(self):
    return self.points.position.shape[0]

  @beartype
  def update_learning_rate(self, lr_scale:float):
    self.points.set_learning_rate(position = self.learning_rates ['position'] * lr_scale)



  def __repr__(self):
    return f"GaussianScene({self.points.position.shape[0]} points)"

  def step(self, visible:torch.Tensor):
    self.points.feature.grad[..., 1:] /= self.config.sh_ratio
    self.points.step()

    self.points.rotation = torch.nn.Parameter(
      F.normalize(self.points.rotation.detach(), dim=1), requires_grad=True)

  def zero_grad(self):
    self.points.zero_grad()



  def split_and_prune(self, keep_mask:torch.Tensor, split_idx:torch.Tensor):

    splits = split_gaussians_uniform(self.gaussians[split_idx], n=2)
    self.points = self.points[keep_mask].append_tensors(splits.to_tensordict())

    return self


  def write_to(self, output_dir:Path, base_filename:str):
    output_dir.mkdir(parents=True, exist_ok=True)
    
    write_gaussians(output_dir / f'{base_filename}.ply', self.gaussians, with_sh=True)

    
  def log(self, logger:Logger, step:int):
    
    point_cloud = PointCloud(
      self.gaussians.position.detach(), 
      sh_to_rgb(self.gaussians.feature[:, :, 0]),
      batch_size=self.gaussians.shape[:1] )
    logger.log_cloud("point_cloud", point_cloud, step=step)
    
    

  @property
  def gaussians(self):
      return Gaussians3D.from_tensordict(self.points.tensors)
      

  def render(self, camera_params:CameraParams, cam_idx:torch.Tensor, 
             **options) -> Rendering:
    
    
    return render_gaussians(self.gaussians, 
                     use_sh        = True,
                     config        = self.raster_config,
                     camera_params = camera_params,
                     **options)
  







from abc import ABCMeta, abstractmethod
import torch

from taichi_splatting import Gaussians3D, Rendering
from taichi_splatting.perspective import CameraParams

from splat_trainer.camera_table.camera_table import CameraTable
from splat_trainer.logger.logger import Logger

  

class GaussianScene(metaclass=ABCMeta):  
  @abstractmethod
  def step(self, visible:torch.Tensor):
    raise NotImplementedError

  @abstractmethod
  def render(self, camera_params:CameraParams, cam_idx:torch.Tensor, **options) -> Rendering:
    raise NotImplementedError

  @abstractmethod
  def split_and_prune(self, keep_mask:torch.Tensor, split_idx:torch.Tensor):
    raise NotImplementedError

  @abstractmethod
  def update_learning_rate(self, lr_scale:float):
    raise NotImplementedError

  @abstractmethod
  def write_to(self, output_dir:str, base_filename:str):
    raise NotImplementedError
  
  @abstractmethod
  def log(self, logger:Logger):
    raise NotImplementedError


class GaussianSceneConfig(metaclass=ABCMeta):


  @abstractmethod
  def from_color_gaussians(self, gaussians:Gaussians3D, camera_table:CameraTable, device:torch.device) -> GaussianScene:
    raise NotImplementedError




import abc
from dataclasses import  dataclass
import math

import torch

class Scheduler(metaclass=abc.ABCMeta):
  @abc.abstractmethod
  def __call__(self, step:int, total_steps:int) -> float:
    pass


@dataclass(kw_only=True, frozen=True)
class ExponentialDecay(Scheduler):
  base_lr: float = 1.0
  final_lr:float = 1/1000.0
  warmup_steps:int = 0

  def __call__(self, step:int, total_steps:int):
    t = min(step / total_steps, 1)
    lr = math.exp(math.log(self.base_lr) * (1 - t) + math.log(self.final_lr * self.base_lr) *  t) 
    
    if step < self.warmup_steps:
      warmup = math.sin(0.5 * math.pi * step / self.warmup_steps)
      lr *= warmup
    
    return lr
  
@dataclass(kw_only=True, frozen=True)
class Uniform(Scheduler):
  base_lr: float = 1.0

  def __call__(self, step:int, total_steps:int):
    return self.base_lr
  






import traceback
import hydra

from omegaconf import OmegaConf
from splat_trainer.util import config

import numpy as np
import torch


config.add_resolvers()


@hydra.main(config_name="config", config_path="../config", version_base="1.2")
def main(cfg):
  train_with_config(cfg)

def train_with_config(cfg):
  import taichi as ti
  from splat_trainer.trainer import Trainer

  torch.set_grad_enabled(False)
  torch.set_printoptions(precision=4, sci_mode=False)
  np.set_printoptions(precision=4, suppress=True)
  torch.set_float32_matmul_precision('high')
  
  print(config.pretty(cfg))
  logger = hydra.utils.instantiate(cfg.logger, _partial_=True)(log_config=OmegaConf.to_container(cfg, resolve=True))
  trainer = None

  try:
    ti.init(arch=ti.cuda, debug=cfg.debug, device_memory_GB=0.1)
    
    train_config = hydra.utils.instantiate(cfg.trainer)
    dataset = hydra.utils.instantiate(cfg.dataset)

    trainer = Trainer(dataset, train_config, logger)

    trainer.train()
    if cfg.wait_exit:
      input("Press Enter to continue...")

  except KeyboardInterrupt:
    pass
  except Exception:
    # print exception and stack trace and exit
     traceback.print_exc()

  if trainer is not None:
    trainer.close()
  
  logger.close()
    
if __name__ == "__main__":
  main()


import torch
import cv2
import numpy as np

def get_cv_colormap(cmap=cv2.COLORMAP_TURBO):
  colormap = np.arange(256, dtype=np.uint8).reshape(1, -1)
  colormap = cv2.applyColorMap(colormap, cmap)
  return torch.from_numpy(colormap.astype(np.uint8)).squeeze(0) 

 
def colorize_depth(color_map, depth, near=0.1):
  depth = depth.clone()
  depth[depth == 0] = torch.inf

  min_depth = torch.clamp(depth, min=near).min()

  inv_depth =  (min_depth / depth).clamp(0, 1)
  inv_depth = (255 * inv_depth).to(torch.int)

  return (color_map[inv_depth])



from functools import partial
from pprint import pformat
import subprocess
from beartype.typing import Dict, List
from beartype import beartype
import torch

from pathlib import Path
from threading import Thread
from queue import Queue

from torch.utils.tensorboard import SummaryWriter
import tabulate

from splat_trainer.logger.histogram import Histogram
from splat_trainer.util.pointcloud import PointCloud

from .logger import Logger


class TensorboardLogger(Logger):
  def __init__(self, log_config:dict, 
               dir: str | None = None, 
               flush_secs:int = 30,
               start_server:bool = False):
    
    if dir is not None:
      Path(dir).mkdir(parents=True, exist_ok=True)

    self.writer = SummaryWriter(dir, flush_secs=flush_secs) 

    self.queue = Queue()
    self.log_thread = Thread(target=self.worker)
    self.log_thread.start()


    if start_server:
      self.process = subprocess.Popen(["tensorboard", "--logdir", self.writer.log_dir]) 
    else:
      self.process = None

    config_str = pformat(log_config)
    self.enqueue(self.writer.add_text, "config", config_str, global_step=0)



  def worker(self):
    item = self.queue.get()
    while item is not None:
      item()
      item = self.queue.get()


  def close(self):
    print("Ending logging thread...")
    self.queue.put(None)
    self.log_thread.join()

    print("Done, closing writer...")

    self.writer.close()

    if self.process is not None:
      self.process.terminate()
    
  @beartype
  def log_evaluations(self, name, rows:List[Dict], step):
    headers = list(rows[0].keys())
    rows = [list(row.values()) for row in rows]
    table = tabulate.tabulate(rows, headers, tablefmt="pipe", floatfmt=".3f")

    self.enqueue(self.writer.add_text, name, table, global_step=step)

  def enqueue(self, func, *args, **kwargs):
    self.queue.put(partial(func, *args, **kwargs))

  @beartype
  def log_values(self, name:str, data:dict, step:int):
    self.enqueue(self.writer.add_scalars, name, data, global_step=step) 

  @beartype
  def log_value(self, name:str, value:float, step:int):
    self.enqueue(self.writer.add_scalar, name, value, global_step=step) 


  @beartype
  def log_image(self, name:str, image:torch.Tensor, caption: str | None = None, step:int = 0):
    self.enqueue(self.writer.add_image, 
                 tag=name, img_tensor=image, 
                 dataformats="HWC",
                 global_step=step)
    
  
  def log_cloud(self, name:str, points:PointCloud, step:int):
    pass # Not supported by tensorboard

  @beartype
  def log_histogram(self, name:str, values:torch.Tensor | Histogram, step:int):
    if isinstance(values, torch.Tensor):
      self.enqueue(self.writer.add_histogram, name, values, global_step=step)
    elif isinstance(values, Histogram):
      self.enqueue(write_histogram, self.writer, name, values, step)  


def write_histogram(writer:SummaryWriter, name:str, hist:Histogram, step:int):
  writer.add_histogram_raw(
    name, 
    min=hist.range[0], max=hist.range[1], 
    num=hist.counts.sum().item(),
    sum=hist.sum,
    sum_squares=hist.sum_squares,
    bucket_limits=hist.bins[1:],
    bucket_counts=hist.counts,  
    global_step=step)


from abc import ABCMeta, abstractmethod
from beartype.typing import  List
import torch

from splat_trainer.logger.histogram import Histogram
from splat_trainer.util.pointcloud import PointCloud

# os.environ["WANDB_SILENT"] = "true"

class Logger(metaclass=ABCMeta):

  @abstractmethod
  def log_evaluations(self, data:List[dict], step:int):
    raise NotImplementedError

  
  @abstractmethod
  def log_image(self, name:str, image:torch.Tensor, step:int, caption:str | None = None):
    raise NotImplementedError
  
  @abstractmethod
  def log_cloud(self, name_str, points:PointCloud, step:int):
    raise NotImplementedError
  
  @abstractmethod
  def log_values(self, name:str, data:dict, step:int):
    raise NotImplementedError
  
  @abstractmethod
  def log_value(self, name:str, value:float, step:int):
    raise NotImplementedError
  
  @abstractmethod
  def log_histogram(self, name:str, values:torch.Tensor | Histogram, step:int):
    raise NotImplementedError

  @abstractmethod
  def close(self):
    raise NotImplementedError
  
  




from functools import partial
from numbers import Number
from beartype.typing import Dict, List
from beartype import beartype
import torch
import wandb  
from pathlib import Path
from threading import Thread
from queue import Queue

from splat_trainer.logger.histogram import Histogram
from splat_trainer.util.pointcloud import PointCloud

from .logger import Logger

class WandbLogger(Logger):
  def __init__(self, project:str | None, log_config:dict, name:str | None=None, dir:str | None = None):
    if dir is not None:
      dir = Path(dir).mkdir(parents=True, exist_ok=True)
    self.run = wandb.init(project=project, name=name, config=log_config, dir=dir, settings=wandb.Settings(start_method='thread'))

    self.queue = Queue()
    self.log_thread = Thread(target=self.worker)
    self.log_thread.start()



  def worker(self):
    item = self.queue.get()
    while item is not None:
      item()
      item = self.queue.get()


  def close(self):
    self.queue.put(None)
    self.log_thread.join()

    self.run.finish(quiet=True)
    

    
  @beartype
  def log_evaluations(self, name, rows:List[Dict], step):
    table = wandb.Table(columns=list(rows[0].keys()))
    for row in rows:
      table.add_data(*row.values())
                        
    self.log_data({name:table}, step=step)


  @beartype
  def log_data(self, data:dict, step:int):
    self.queue.put(partial(self.run.log, data, step=step) )

  @beartype
  def log_value(self, name:str, value:Number, step:int):
    self.log_data({name:value}, step=step)

  @beartype
  def log_values(self, name:str, data:dict, step:int):
    self.log_data({f"{name}/{k}":value for k, value in data.items()}, step=step)


  @beartype
  def log_image(self, name:str, image:torch.Tensor, caption:str | None = None, step:int = 0):
    
    def log():
      nonlocal image, step
      
      image = (image * 255).to(torch.uint8).cpu().numpy()
      image = wandb.Image(image, mode="RGB", caption=caption, file_type="jpg")
      self.run.log({name : image}, step=step)

    self.queue.put(log)


  def log_cloud(self, name:str, points:PointCloud, step:int):

    def log():
      nonlocal points, step
      
      data = torch.cat([points.points, points.colors * 255], dim=1).cpu().numpy()

      image = wandb.Object3D(data)
      self.run.log({name : image}, step=step)

    self.queue.put(log)


  @beartype
  def log_histogram(self, name:str, values:torch.Tensor | Histogram, step:int):
    def log():
      try:
        if isinstance(values, Histogram):
          hist = wandb.Histogram(np_histogram=
            (values.counts.cpu().numpy(), values.bins.cpu().numpy()))
          
          self.run.log({name:hist}, step=step)

        elif isinstance(values, torch.Tensor):
          self.run.log({name:wandb.Histogram(values.cpu().numpy())}, step=step)
      except Exception as e:
        print(f"Error logging histogram {name}: {e}")

    self.queue.put(log)



from .controller import ControllerConfig, Controller

from .target_controller import TargetConfig, TargetController
from .threshold_controller import ThresholdConfig, ThresholdController

__all__ = ['ControllerConfig', 'Controller', 'TargetConfig', 'TargetController', 'ThresholdConfig', 'ThresholdController']


import abc
from dataclasses import dataclass
import math
from beartype import beartype
import torch
import torch.nn.functional as F

from taichi_splatting.perspective import CameraParams



class ImageScaler(metaclass=abc.ABCMeta):
  @abc.abstractmethod
  def __call__(self, image:torch.Tensor, camera_params:CameraParams, current_factor:float):
    pass

  @abc.abstractmethod
  def update(self, step:int) -> float:
    pass
    
    

def interpolate_channels_last(image, scale, mode='bilinear'):
    image = image.unsqueeze(0).permute(0, 3, 1, 2).to(memory_format=torch.channels_last)
    image = F.interpolate(image, scale_factor=scale, mode=mode)
    return image.permute(0, 2, 3, 1).squeeze(0).to(memory_format=torch.contiguous_format)

@dataclass(kw_only=True, frozen=True)
class ExponentialScaler(ImageScaler):
  initial_scale: float = 0.25
  final_scale: float = 1.0

  steps: int = 20000
  scale_mode:str = 'bilinear'

  
  def update(self, step:int) -> float:
    t = min(step / self.steps, 1)
    return math.exp(math.log(self.initial_scale) * (1 - t) +  t * math.log(self.final_scale)) 


  @beartype
  def __call__(self, image:torch.Tensor, camera_params:CameraParams, current_factor:float):
    camera_params = camera_params.scale_image(current_factor)
    image = interpolate_channels_last(image, current_factor, mode=self.scale_mode)
    return image, camera_params
  




@dataclass(kw_only=True, frozen=True)
class NullScaler:

  def update(self, step:int) -> float:
    return 1.0

  @beartype
  def __call__(self, image:torch.Tensor, camera_params:CameraParams, current_factor:float):
    return image, camera_params



from dataclasses import  dataclass
import math
from beartype.typing import Optional
from taichi_splatting import Rendering
from tensordict import tensorclass
import torch

from splat_trainer.logger.logger import Logger
from .controller import Controller, ControllerConfig
from splat_trainer.scene import GaussianScene



@tensorclass
class PointStatistics:
  point_grad : torch.Tensor  # (N, 2) - accumulated split heuristics
  visible : torch.Tensor  # (N, ) - number of times the point was in view

  @staticmethod
  def zeros(batch_size, device:Optional[torch.device] = None):
    return PointStatistics(
      point_grad=torch.zeros(batch_size, dtype=torch.float32, device=device),
      visible=torch.zeros(batch_size, dtype=torch.float32, device=device),
      batch_size=(batch_size,)
    )



@dataclass
class ThresholdConfig(ControllerConfig):

  enabled:bool = True
  grad_threshold:float = 0.0002
  min_split_size:float = 0.001

  min_opacity:float = 0.01

  # min number of times a point must be visible recently to be considered for splitting/cloning
  min_visibility:int = 40 

  def make_controller(self, scene:GaussianScene, 
               densify_interval:int, total_steps:int):
    return ThresholdController(self, scene,  densify_interval, total_steps)


class ThresholdController(Controller):
  def __init__(self, config:ThresholdConfig, 
               scene:GaussianScene,
               densify_interval:int, 
               total_steps:int):
    
    self.config = config
    self.scene = scene

    self.densify_interval = densify_interval
    self.total_steps = total_steps
    self.points = PointStatistics.zeros(scene.num_points, device=scene.device)
 

  def log_histograms(self, logger:Logger, step:int):
    point_grad = self.points.point_grad

    logger.log_histogram("points/log_point_grad", 
        point_grad[(point_grad > 0) & point_grad.isfinite()].log(), step)
    logger.log_histogram("points/visible", self.points.visible, step)


  def find_split_prune_indexes(self, step:int):
      config = self.config  

      grad_mean = self.points.point_grad / self.points.visible
      vis_mask = (self.points.visible > config.min_visibility)

      split_mask = (grad_mean > config.grad_threshold) & vis_mask
      prune_mask = self.scene.gaussians.alpha_logit.squeeze(1) < math.log(config.min_opacity)
      keep_mask = ~prune_mask & ~split_mask

      split_idx = split_mask.nonzero(as_tuple=False).squeeze(1)

      counts = dict(n=self.points.batch_size[0], 
          visible=vis_mask.sum().item(),
          split=split_idx.shape[0],
          prune=prune_mask.sum().item())
      
      return keep_mask, split_idx, counts     


  def densify_and_prune(self, step:int):

    keep_mask, split_idx, counts = self.find_split_prune_indexes(step)

    if self.config.enabled:
      self.scene.split_and_prune(keep_mask, split_idx)
      self.points = self.points[keep_mask]

      new_points = PointStatistics.zeros(split_idx.shape[0] * 2, device=self.scene.device)
      self.points = torch.cat([self.points, new_points], dim=0)

    return counts    




  def add_rendering(self, rendering:Rendering): 
    idx = rendering.points_in_view

    longest_side = max(*rendering.image_size)
    point_grad = rendering.split_heuristics[:, 1] * 0.5 * longest_side

    visible_mask = (point_grad > 0)

    self.points.point_grad[idx] += point_grad
    self.points.visible[idx] += visible_mask

    return (idx[visible_mask], idx)






from dataclasses import dataclass
from pathlib import Path
from beartype import beartype
from beartype.typing import Iterator, Tuple, List

import torch

import numpy as np
from splat_trainer.camera_table.camera_table import MultiCameraTable
from splat_trainer.dataset.colmap.loading import load_images
from splat_trainer.dataset.dataset import CameraView, Dataset

from splat_trainer.util.misc import split_stride
from splat_trainer.util.pointcloud import PointCloud

import pycolmap 



class COLMAPDataset(Dataset):
  def __init__(self, base_path:str,
        model_dir:str = "sparse/0",
        image_dir:str = "images",                
        image_scale:float=1.0,
        val_stride:int=10,
        depth_range:Tuple[float, float] = (0.1, 100.0)):

    self.image_scale = image_scale
    self.depth_range = depth_range

    self.base_path = base_path
    model_path = Path(base_path) / model_dir

    self.reconstruction = pycolmap.Reconstruction(str(model_path))
    self.projections = []
    camera_idx = {}


    for i, (k, camera) in enumerate(self.reconstruction.cameras.items()):
      camera.rescale(image_scale)
      m = camera.calibration_matrix()
      self.projections.append(m)
      camera_idx[k] = i
      print(f"Camera {k}@{camera.width}x{camera.height} fx={m[0, 0]:.2f} fy={m[1, 1]:.2f} cx={m[0, 2]:.2f} cy={m[1, 2]:.2f}")


    def image_info(image:pycolmap.Image) -> Tuple[np.array, str, int]:
      camera_t_world = np.eye(4)
      camera_t_world[:3, :4] =  image.cam_from_world.matrix()    

      return (camera_t_world, image.name, camera_idx[image.camera_id])
    
    self.num_cameras = len(self.reconstruction.images)
    self.camera_t_world, self.image_names, self.camera_idx = zip(
      *[image_info(image) for image in self.reconstruction.images.values()])
    
    images = load_images(list(self.image_names), Path(base_path) / image_dir, image_scale=image_scale)
    cameras = [CameraImage(filename, torch.from_numpy(image).pin_memory(), i) 
               for i, (filename, image) in enumerate(zip(self.image_names, images))]  
  
    # Evenly distribute validation images
    self.train_cameras, self.val_cameras = split_stride(cameras, val_stride)

  def __repr__(self) -> str:
    return f"COLMAPDataset({self.base_path}, image_scale={self.image_scale})"

  def train(self, shuffle=True) -> Iterator[CameraView]:
    return Images(self.train_cameras, shuffle=shuffle)
  
  def val(self) -> Iterator[CameraView]:
    return Images(self.val_cameras)


  def camera_table(self) -> MultiCameraTable:
    return MultiCameraTable(
      camera_t_world = torch.tensor(np.array(self.camera_t_world), dtype=torch.float32),
      projection = torch.tensor(np.array(self.projections), dtype=torch.float32),
      camera_idx = torch.tensor(self.camera_idx, dtype=torch.long))


  def pointcloud(self) -> PointCloud:  
    xyz = np.array([p.xyz for p in self.reconstruction.points3D.values()])
    colors = np.array([p.color for p in self.reconstruction.points3D.values()])

    return PointCloud(torch.tensor(xyz, dtype=torch.float32), 
                      torch.tensor(colors, dtype=torch.float32) / 255.0,
                      batch_size=(len(xyz),))
    





@dataclass
class CameraImage:
  filename:str
  image:torch.Tensor
  image_id:int

class Images(torch.utils.data.Dataset):
  @beartype
  def __init__(self, camera_images:List[CameraImage], shuffle:bool=False):
    self.camera_images = camera_images
    self.shuffle = shuffle

  def __len__(self):
      return len(self.camera_images)

  def __getitem__(self, index) -> CameraView:
    camera_image:CameraImage = self.camera_images[index]

    idx = torch.tensor([camera_image.image_id], dtype=torch.long).pin_memory()
    return camera_image.filename, camera_image.image, idx
     
  def __iter__(self) -> Iterator[CameraView]:
    order = torch.randperm(len(self)) if self.shuffle else torch.arange(len(self))
    for idx in order:
      yield self[idx]  




import numpy as np
import torch

def strided_indexes(subset:int, total:int):
  if subset > 0:
    stride = max(total // subset, 1)
    return torch.arange(0, total, stride).to(torch.long)
  else:
    return torch.tensor([])


def split_stride(images, stride=0):
  assert stride == 0 or stride > 1, f"val_stride {stride}, must be zero, or greater than 1"

  val_cameras = [camera for i, camera in enumerate(images) if i % stride == 0] if stride > 0 else []
  train_cameras = [camera for camera in images if camera not in val_cameras]

  return train_cameras, val_cameras



def sigmoid(x):
  return 1 / (1 + np.exp(-x))

def inverse_sigmoid(x):
  return np.log(x / (1 - x))


sh0 = 0.282094791773878

def rgb_to_sh(rgb):
    return (rgb - 0.5) / sh0

def sh_to_rgb(sh):
    return sh * sh0 + 0.5


class CudaTimer:
  def __init__(self):
    self.start = torch.cuda.Event(enable_timing = True)
    self.end = torch.cuda.Event(enable_timing = True)

  def __enter__(self):
    self.start.record()

  def __exit__(self, *args):
    self.end.record()

  def ellapsed(self):
    return self.start.elapsed_time(self.end)


from taichi_splatting.tests.random_data import random_3d_gaussians, random_camera

from taichi_splatting import render_gaussians
from splat_trainer.gaussians import split

import torch

import cv2
import taichi as ti


def show_image(image):
  image = (image * 255).to(torch.uint8).cpu().numpy()
  cv2.imshow("rendering", image)
  return cv2.waitKey(0)


def main():
  ti.init(arch=ti.cuda, debug=True)

  while True:
    camera = random_camera(image_size=(640, 480))
    gaussians = random_3d_gaussians(5, camera, alpha_range=(1.0, 1.0), scale_factor=0.2)

    device = torch.device("cuda:0")
    camera, gaussians = camera.to(device), gaussians.to(device)

    image = render_gaussians(gaussians, camera_params=camera).image
    show_image(image)

    split_gaussians = split.split_gaussians(gaussians, n=2)

    image = render_gaussians(split_gaussians, camera_params=camera).image
    show_image(image)

if __name__ == "__main__":
  main()


from functools import partial
from multiprocessing import cpu_count, get_logger
from multiprocessing.pool import ThreadPool
from pathlib import Path
from beartype.typing import List


from beartype import beartype
import cv2
from tqdm import tqdm
import traceback

# Shortcut to multiprocessing's logger
def error(msg, *args):
    return get_logger().error(msg, *args)

class LogExceptions(object):
    def __init__(self, callable):
        self.__callable = callable

    def __call__(self, *args, **kwargs):
        try:
            result = self.__callable(*args, **kwargs)

        except Exception as e:
            # Here we add some debugging help. If multiprocessing's
            # debugging is on, it will arrange to log the traceback
            error(traceback.format_exc())
            # Re-raise the original exception so the Pool worker can
            # clean up
            raise

        # It was fine, give a normal answer
        return result


def parmap_list(f, xs, j=cpu_count() // 2, chunksize=1, pool=ThreadPool, progress=tqdm):

  with pool(processes=j) as pool:
    iter = pool.imap(LogExceptions(f), xs, chunksize=chunksize)
    
    if progress is not None:
      iter = progress(iter, total=len(xs))

    return list(iter)


@beartype
def load_image(filename:Path, image_scale:float=1.0):
  assert filename.is_file(), f"load_image: file {filename} does not exist"

  image = cv2.imread(str(filename), cv2.IMREAD_COLOR)

  if image_scale != 1.0:
    h, w, _ = image.shape
    dsize = (int(w * image_scale), int(h * image_scale))

    image = cv2.resize(image, dsize=dsize, fx=image_scale, fy=image_scale, interpolation=cv2.INTER_AREA)

  image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

  assert image is not None, f"load_image: could not read {filename}"
  return image

@beartype
def load_images(filenames:List[str], base_path:Path, image_scale:float, **map_options):
  return parmap_list(partial(load_image, image_scale=image_scale), [base_path / f for f in filenames], **map_options)  



from tensordict import tensorclass
from pathlib import Path

import torch


import numpy as np
import pypcd4
import plyfile



@tensorclass 
class PointCloud:
  points : torch.Tensor # (N, 3)
  colors : torch.Tensor # (N, 3)


  @staticmethod
  def from_numpy(xyz:np.ndarray, rgb:np.ndarray) -> 'PointCloud':
    if rgb.dtype == np.uint8:
      rgb = rgb.astype(np.float32) / 255.0


    return PointCloud(
        points = torch.from_numpy(xyz.astype(np.float32)), 
        colors = torch.from_numpy(rgb.astype(np.float32)),
        batch_size = (xyz.shape[0],))

  def show(self):
    import open3d as o3d
          
    pcl = o3d.geometry.PointCloud()
    pcl.points = o3d.utility.Vector3dVector(self.points.cpu().numpy())
    pcl.colors = o3d.utility.Vector3dVector(self.colors.cpu().numpy())

    o3d.visualization.draw_geometries([pcl])

  
  
  @staticmethod
  def load(filename:str | Path):
    filename = Path(filename)

    if filename.suffix == ".pcd": 
      cloud = pypcd4.PointCloud.from_path(filename)
      if 'rgb' in cloud.fields:
        rgb = cloud.numpy(('rgb',))
        rgb = rgb.view(np.uint8).reshape(-1, 4)
        rgb = rgb[:, :3].astype(np.float32) / 255.0

      elif 'r' in cloud.fields:
        rgb = cloud.numpy(('red', 'green', 'blue'))
      else:
        rgb = np.ones((cloud.size, 3), dtype=np.uint8)
        
      return PointCloud.from_numpy(cloud.numpy(("x", "y", "z")), rgb)

    elif filename.suffix == ".ply":
      ply = plyfile.PlyData.read(filename) 
      v = ply['vertex']
      xyz = np.stack([v['x'], v['y'], v['z']], axis=1)
      rgb = np.stack([v['red'], v['green'], v['blue']], axis=1)

      return PointCloud.from_numpy(xyz, rgb)  
    else:
      raise ValueError(f"Unknown file type {filename.suffix}")


from dataclasses import dataclass
import math
import os
from pathlib import Path

from beartype.typing import Optional
import numpy as np
import torch

import torch.nn.functional as F
from torchmetrics.image  import MultiScaleStructuralSimilarityIndexMeasure

 
from taichi_splatting import Rendering, perspective
from tqdm import tqdm

from splat_trainer.dataset import Dataset
from splat_trainer.image_scaler import ImageScaler, NullScaler
from splat_trainer.gaussians.loading import from_pointcloud

from splat_trainer.logger import Logger
from splat_trainer.logger.histogram import Histogram

from splat_trainer.scene.sh_scene import  GaussianSceneConfig
from splat_trainer.scheduler import Scheduler, Uniform
from splat_trainer.util.colorize import colorize_depth, get_cv_colormap
from splat_trainer.util.containers import transpose_rows
from splat_trainer.util.misc import CudaTimer, strided_indexes

from splat_trainer.controller import ControllerConfig

@dataclass(kw_only=True)
class TrainConfig:
  output_path: str
  device: str
  steps: int 
  scene: GaussianSceneConfig
  controller: ControllerConfig

  load_model: Optional[str] = None

  num_neighbors:int   = 3
  initial_point_scale:float = 0.5
  initial_alpha:float = 0.5 

  densify_interval: int = 50
  update_steps: int = 10

  eval_steps: int = 1000
  num_logged_images: int = 5
  log_interval: int = 20

  ssim_weight: float = 0.0
  ssim_scale: float = 0.5

  blur_cov: float = 0.3

  lr_scheduler: Scheduler = Uniform()
  image_scaler: ImageScaler = NullScaler()
  

class Trainer:
  def __init__(self, dataset:Dataset, config:TrainConfig, logger:Logger):

    self.device = torch.device(config.device)

    self.dataset = dataset
    self.config = config
    self.step = 0
    self.logger = logger
    self.image_scale = self.config.image_scaler.update(0)

    self.blur_cov = config.blur_cov
    self.output_path = Path(config.output_path or os.getcwd())

    self.ssim = MultiScaleStructuralSimilarityIndexMeasure(
        data_range=1.0, kernel_size=11).to(self.device)

    self.camera_table = dataset.camera_table()
    self.camera_table.to(self.device)
    self.camera_table.requires_grad_(False)
    

    print(f"Initializing model from {dataset}")
    initial_gaussians = from_pointcloud(dataset.pointcloud(), 
                                        initial_scale=config.initial_point_scale,
                                        initial_alpha=config.initial_alpha,
                                        num_neighbors=config.num_neighbors)
    
    self.scene = config.scene.from_color_gaussians(initial_gaussians, self.camera_table, self.device)
    print(self.scene)

    self.render_timers = [CudaTimer() for _ in range(self.config.log_interval)]

    self.controller = config.controller.make_controller(
      self.scene,  config.densify_interval, config.steps)
    
    self.color_map = get_cv_colormap().to(self.device)

    self.pbar = None
    

  def camera_params(self, cam_idx:torch.Tensor, image:torch.Tensor):
        near, far = self.dataset.depth_range
        camera_t_world, image_t_camera = self.camera_table(cam_idx)

        return perspective.CameraParams(
            T_camera_world=camera_t_world.squeeze(0),
            T_image_camera=image_t_camera.squeeze(0),
            image_size=(image.shape[1], image.shape[0]),
            near_plane=near,
            far_plane=far,
            blur_cov=self.blur_cov
        ).to(self.device, dtype=torch.float32)


  def log_image(self, name, image, caption=None):
    return self.logger.log_image(name, image, caption=caption, step=self.step)
  

  def log_value(self, name, value):
    return self.logger.log_value(name, value, step=self.step) 

  def log_values(self, name, values):
    return self.logger.log_values(name, values, step=self.step)  

  def log_histogram(self, name, values):
    return self.logger.log_histogram(name, values, step=self.step)



  def evaluate_dataset(self, name, data, log_count:int=0):
    if len(data) == 0:
      return {}

    rows = []
    radius_hist = Histogram.empty(range=(-1, 3), num_bins=20, device=self.device) 
    log_indexes = strided_indexes(log_count, len(data)) 

    pbar = tqdm(total=len(data), desc=f"rendering {name}", leave=False)
    for i, (filename, camera_params, cam_idx, image) in enumerate(self.iter_data(data)):

      rendering = self.scene.render(camera_params, cam_idx, compute_radii=True, render_depth=True)
      psnr = compute_psnr(rendering.image, image)
      l1 = torch.nn.functional.l1_loss(rendering.image, image)

      radius_hist = radius_hist.append(rendering.radii.log() / math.log(10.0), trim=False)

      if i in log_indexes:
        image_id = filename.replace("/", "_")
        self.log_image(f"{name}_images/{image_id}/render", rendering.image, caption=f"{filename} PSNR={psnr:.2f} L1={l1:.2f}")
        self.log_image(f"{name}_images/{image_id}/depth", 
            colorize_depth(self.color_map, rendering.depth, 0.1), caption=filename)

        if self.step == 0:
          self.log_image(f"{name}_images/{image_id}/image", image, caption=filename)
      
      eval = dict(filename=filename, psnr = psnr.item(), l1 = l1.item())
      rows.append(eval)
      
      pbar.update(1)
      pbar.set_postfix(psnr=f"{psnr.item():.2f}", l1=f"{l1.item():.4f}")

    self.logger.log_evaluations(f"{name}/evals", rows, step=self.step)
    totals = transpose_rows(rows)
    mean_l1, mean_psnr = np.mean(totals['l1']), np.mean(totals['psnr'])

    self.log_value(f"{name}/psnr", mean_psnr) 
    self.log_value(f"{name}/l1", mean_l1) 

    self.log_histogram(f"{name}/psnr_hist", torch.tensor(totals['psnr']))
    self.log_histogram(f"{name}/radius_hist", radius_hist)

    return {f"{name}_psnr":mean_psnr}


  def evaluate(self):
    n_logged = self.config.num_logged_images

    train = self.evaluate_dataset("train", self.dataset.train(shuffle=False), log_count=n_logged)
    val = self.evaluate_dataset("val", self.dataset.val(), log_count=n_logged)

    self.scene.write_to(self.output_path / "point_cloud" , f"model_{self.step}")
    # self.scene.log(self.logger, self.step)

    return {**train, **val}
  


  def iter_train(self):
    while True:
      train = self.iter_data(self.dataset.train())
      yield from train

  

  def iter_data(self, iter):
    for filename, image, image_idx in iter:
      image, image_idx = [x.to(self.device, non_blocking=True) 
                    for x in (image, image_idx)]
      
      image = image.to(dtype=torch.float) / 255.0
      camera_params = self.camera_params(image_idx, image)

      yield filename, camera_params, image_idx.squeeze(0), image

  @torch.compile()
  def compute_ssim(self, image:torch.Tensor, ref:torch.Tensor, scale:float=1.0):
      image1 = ref.unsqueeze(0).permute(0, 3, 1, 2).to(memory_format=torch.channels_last)
      image2 = image.unsqueeze(0).permute(0, 3, 1, 2).to(memory_format=torch.channels_last)

      return 1. - self.ssim(F.interpolate(image1, scale_factor=scale), F.interpolate(image2, scale_factor=scale))
      

  def losses(self, rendering:Rendering, image):
    l1 = torch.nn.functional.l1_loss(rendering.image, image)
    
    if self.config.ssim_weight > 0:  
      ssim = self.compute_ssim(rendering.image, image, scale=self.config.ssim_scale)
      loss = l1 + ssim * self.config.ssim_weight 
      return loss, dict(l1=l1.item(), ssim=ssim.item())

    else:
      return l1, dict(l1=l1.item())


  def training_step(self, filename, camera_params, image_idx, image, timer):
    image, camera_params = self.config.image_scaler(image, camera_params, self.image_scale)

    with timer:
      rendering = self.scene.render(camera_params, image_idx, compute_split_heuristics=True)

      loss, losses = self.losses(rendering, image)
      loss.backward()

    (visible, in_view) =  self.controller.add_rendering(rendering)
    self.scene.step(visible)

    self.scene.zero_grad()

    self.step += 1
    return dict(**losses, visible=visible.shape[0], in_view=in_view.shape[0])


  def train(self):
    print(f"Writing to model path {os.getcwd()}")


    self.pbar = tqdm(total=self.config.steps, desc="training")
    self.step = 0
    since_densify = 0
    densify_metrics = dict(n = self.scene.num_points)

    iter_train = self.iter_train()
    step_timer = CudaTimer()

    while self.step < self.config.steps:


      if self.step % self.config.eval_steps == 0:
          eval_metrics = self.evaluate()
          self.image_scale = self.config.image_scaler.update(self.step)

          lr_scale = self.config.lr_scheduler(self.step, self.config.steps)
          self.scene.update_learning_rate(lr_scale)

          self.log_values("train", dict(lr_scale=lr_scale, blur_cov=self.blur_cov, image_scale=self.image_scale))

      if since_densify >= self.config.densify_interval:
        self.controller.log_histograms(self.logger, self.step)
        densify_metrics = self.controller.densify_and_prune(self.step)

        self.log_values("densify", densify_metrics)
        since_densify = 0
      
      torch.cuda.empty_cache()

      with torch.enable_grad():
        with step_timer:
          steps = [self.training_step(*next(iter_train), timer=timer) 
                  for timer in self.render_timers]

      since_densify += len(steps)

      if self.step % self.config.log_interval  == 0:
        steps = transpose_rows(steps)

        self.pbar.update(self.config.log_interval)

        means = {k:np.mean(v) for k, v in steps.items()}
        metrics = {k:f"{means[k]:.4f}" for k in ['l1', 'ssim', 'reg'] if k in means}

        self.pbar.set_postfix(**metrics, **eval_metrics, **densify_metrics)        
        self.log_values("train", means)

        # skip first step as it will include compiling kernels
        if self.step > self.config.log_interval:
          torch.cuda.synchronize()

          self.log_values("timer", 
                  dict(step_ms=step_timer.ellapsed() / self.config.update_steps,
                  render=sum([timer.ellapsed() for timer in self.render_timers]) / self.config.update_steps
                ))

    eval_metrics = self.evaluate()
    self.pbar.set_postfix(**metrics, **eval_metrics)        
    self.pbar.close()


    return (eval_metrics.get('eval_train_psnr', 0.0) 
            + eval_metrics.get('val_psnr', 0.0)) / 2.0
    

  def close(self):
    if self.pbar is not None:
      self.pbar.close()


def compute_psnr(a, b):
  return 10 * torch.log10(1 / torch.nn.functional.mse_loss(a, b))  



import math
from numbers import Number
import torch


class Histogram:
  """ A histogram class for torch tensors,
  can be added with other histograms which have the same bins """

  def __init__(self,
               values,
               range=(0, 1),
               num_bins=10,
               trim=True):
    assert len(range) == 2

    self.range = range
    lower, upper = self.range

    bin_indexes = (values - lower) * num_bins / (upper - lower)
    bin_indexes = bin_indexes.long()

    if trim:
      valid = (bin_indexes >= 0) & (bin_indexes < num_bins)

      values = values[valid]
      bin_indexes = bin_indexes[valid]

    bin_indexes.clamp_(0, num_bins - 1)

    self.sum = values.sum().item()
    self.sum_squares = values.norm(2).item()
    self.counts = torch.bincount(bin_indexes, minlength=num_bins)

  @staticmethod
  def empty(range, num_bins, device:torch.device):
    return Histogram(torch.zeros((0,), device=device), range, num_bins)
  
  @property
  def device(self):
    return self.counts.device

  def __repr__(self):
    return self.counts.tolist().__repr__()

  @property
  def bins(self):
    lower, upper = self.range
    d = (upper - lower) / self.counts.size(0)

    return torch.tensor(
        [lower + i * d for i in range(0,
                                      self.counts.size(0) + 1)])
  @property
  def bin_width(self):
    lower, upper = self.range
    return (upper - lower) / self.counts.size(0)
  
  @property
  def bin_ranges(self):
    bins = self.bins()
    return torch.stack([bins[:-1], bins[1:]], dim=1)


  def __add__(self, other):
    assert isinstance(other, Histogram)
    assert other.counts.size(0) == self.counts.size(0) and (
        other.range == self.range), 'mismatched histogram sizes'

    total = Histogram.empty(range=self.range, num_bins=self.counts.size(0), device=self.device)
    total.sum = self.sum + other.sum
    total.sum_squares = self.sum_squares + other.sum_squares
    total.counts = self.counts + other.counts

    return total
  
  def append(self, values:torch.Tensor, trim=True):
    return self + Histogram(values, 
        range=self.range, num_bins=self.counts.size(0), trim=trim)

  def __truediv__(self, x):
    assert isinstance(x, Number)

    total = Histogram(range=self.range, num_bins=self.counts.size(0))
    total.sum = self.sum / x
    total.sum_squares = self.sum_squares / x
    total.counts = self.counts / x

    return total

  @property
  def std(self):

    n = self.counts.sum().item()
    if n > 1:
      sum_squares = self.sum_squares - (self.sum * self.sum / n)
      var = max(0, sum_squares / (n - 1))

      return math.sqrt(var)
    else:
      return 0

  @property
  def mean(self):
    n = self.counts.sum().item()
    if n > 0:
      return self.sum / self.counts.sum().item()
    else:
      return 0




from dataclasses import dataclass
from beartype.typing import Dict, Iterator, List

import cv2
import torch

from camera_geometry import FrameSet, Camera
from camera_geometry.scan.views import load_frames_with, Undistortion

from beartype import beartype
import numpy as np

from splat_trainer.dataset import CameraView

@dataclass
class CameraImage:
   camera : Camera
   image : torch.Tensor
   frame_id: int
   camera_id: int
   filename: str


def concat_lists(xs):
  return [x for x in xs for x in x]


@beartype
def preload_images(scan:FrameSet, undistorted:Dict[str, Camera]) -> List[CameraImage]:
  undistortions = {k:Undistortion(camera, undistorted[k]) 
    for k, camera in scan.cameras.items() }

  camera_names = list(undistortions.keys())

  @beartype
  def load(undistortion: Undistortion, rig_pose:np.ndarray, camera_name:str, frame_index:int):
    image_file = scan.image_sets.rgb[frame_index][camera_name]
    image = scan.loader.rgb.load_image(image_file)
    image = undistortion.undistort(image)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    camera_id = camera_names.index(camera_name)
    return CameraImage(
        camera=undistortion.undistorted.transform(rig_pose),
        image=torch.from_numpy(image).pin_memory(),
        frame_id=frame_index,
        camera_id=camera_id,
        filename=image_file
    )

  frames = load_frames_with(scan, undistortions, load)
  return concat_lists(frames)

class PreloadedImages(torch.utils.data.Dataset):
  @beartype
  def __init__(self, camera_images:List[CameraImage], shuffle:bool=False):
    self.camera_images = camera_images
    self.shuffle = shuffle

  def __len__(self):
      return len(self.camera_images)

  def __getitem__(self, index) -> CameraView:
    camera_image:CameraImage = self.camera_images[index]

    idx = torch.tensor([camera_image.frame_id, camera_image.camera_id], dtype=torch.long).pin_memory()
    return camera_image.filename, camera_image.image, idx
     
  def __iter__(self) -> Iterator[CameraView]:
    order = torch.randperm(len(self)) if self.shuffle else torch.arange(len(self))
    for idx in order:
      yield self[idx]  

    





from dataclasses import  dataclass
from beartype.typing import Optional
import numpy as np
from taichi_splatting import Rendering
from tensordict import tensorclass
import torch

from splat_trainer.logger.logger import Logger
from .controller import Controller, ControllerConfig
from splat_trainer.scene import GaussianScene



@tensorclass
class PointStatistics:
  split_score : torch.Tensor  # (N, ) - averaged split score
  prune_cost : torch.Tensor  # (N, ) - max prune cost

  visible : torch.Tensor  # (N, ) - number of times the point was visible
  in_view : torch.Tensor  # (N, ) - number of times the point was in the view volume
  

  @staticmethod
  def zeros(batch_size, device:Optional[torch.device] = None):
    return PointStatistics(
      split_score=torch.zeros(batch_size, dtype=torch.float32, device=device),
      prune_cost=torch.zeros(batch_size, dtype=torch.float32, device=device),

      visible=torch.zeros(batch_size, dtype=torch.int16, device=device),
      in_view=torch.zeros(batch_size, dtype=torch.int16, device=device),
      batch_size=(batch_size,)
    )
    



@dataclass
class TargetConfig(ControllerConfig):

  # target point cloud size - if None then optimize for the current size
  target_count:Optional[int] = None

  # base rate (relative to count) to split points every 100 iterations
  split_rate:float = 0.05

  # max ratio of points to split/prune
  max_ratio:float = 4.0

  # min number of times a point must be visible recently to be considered for splitting/pruning
  min_visibility:int = 20 

  def make_controller(self, scene:GaussianScene, 
               densify_interval:int, total_steps:int):
    return TargetController(self, scene,  densify_interval, total_steps)


class TargetController(Controller):
  def __init__(self, config:TargetConfig, 
               scene:GaussianScene,
               densify_interval:int, 
               total_steps:int):
    
    self.config = config
    self.scene = scene

    self.densify_interval = densify_interval
    self.total_steps = total_steps

    self.target_count = config.target_count or scene.num_points
    self.points = PointStatistics.zeros(scene.num_points, device=scene.device)
       
    self.splits_per_densify = (1 + config.split_rate) ** (densify_interval / 100) - 1.0
    self.ema_alpha  = (0.1 ** (2.0 / config.min_visibility))



  def log_histograms(self, logger:Logger, step:int):

    split_score, prune_cost = self.points.split_score.log(), self.points.prune_cost.log()

    logger.log_histogram("points/log_split_score", split_score[split_score.isfinite()], step)
    logger.log_histogram("points/log_prune_cost",  prune_cost[prune_cost.isfinite()], step)
    logger.log_histogram("points/visible", self.points.visible, step)


  def find_split_prune_indexes(self, step:int):
      config = self.config  

      split_ratio = np.clip(self.target_count / self.scene.num_points, 
                              a_min=1/config.max_ratio, a_max=config.max_ratio)
        
      t = max(0, (2 * step / self.total_steps) - 1)

      candidates = torch.nonzero(self.points.visible >= config.min_visibility).squeeze(1)
      points = self.points[candidates]


      if candidates.shape[0] == 0:
        splittable, pruneable = [
          torch.zeros(0, dtype=torch.bool, device=candidates.device) for _ in range(2)]

      else:
        # quadratic decay
        quantile = self.splits_per_densify * (1 - t)**2

        split_thresh = torch.quantile(points.split_score, 1 - (quantile * split_ratio))
        prune_thresh = torch.quantile(points.prune_cost, quantile * 1/split_ratio )

        pruneable = (points.prune_cost <= prune_thresh) 
        splittable = (points.split_score > split_thresh) & ~pruneable
      
      split_idx, prune_idx =  candidates[splittable], candidates[pruneable]

      counts = dict(n=self.points.batch_size[0], 
              visible=candidates.shape[0],
              split=split_idx.shape[0],
              prune=prune_idx.shape[0])
         
      return split_idx, prune_idx, counts     


  def densify_and_prune(self, step:int):
    split_idx, prune_idx, counts = self.find_split_prune_indexes(step)

    keep_mask = torch.ones(self.points.batch_size[0], dtype=torch.bool, device=self.scene.device)
    keep_mask[prune_idx] = False
    keep_mask[split_idx] = False

    # keep_mask[self.points.age > 100 & (self.points.visible < 1)] = False

    self.scene.split_and_prune(keep_mask, split_idx)
    self.points = self.points[keep_mask]

    new_points = PointStatistics.zeros(split_idx.shape[0] * 2, device=self.scene.device)
    self.points = torch.cat([self.points, new_points], dim=0)
    return counts    


  def add_rendering(self, rendering:Rendering): 
    idx = rendering.points_in_view
    split_score, prune_cost = rendering.split_heuristics.unbind(1)

    vis_mask = prune_cost > 0
    vis_idx = idx[vis_mask]

    points = self.points

    # weight = torch.where(self.points.in_view[vis_idx] > 0, self.ema_alpha, 1.0)
    # points.split_score[vis_idx] = (1 - weight) * points.split_score[vis_idx] + weight * split_score[vis_mask]

    points.prune_cost[idx] = torch.maximum(points.prune_cost[idx]  * self.ema_alpha, prune_cost) 
    points.split_score[idx] = torch.maximum(points.split_score[idx] * self.ema_alpha, split_score )  


    points.in_view[idx] += 1
    points.visible[vis_idx] += 1

    return (vis_idx, idx)






from os import path
from omegaconf import OmegaConf 


def add_resolvers():
    OmegaConf.register_new_resolver("dirname", path.dirname)
    OmegaConf.register_new_resolver("basename", path.basename)
    OmegaConf.register_new_resolver("without_ext", lambda x: path.splitext(x)[0])
    OmegaConf.register_new_resolver("ext", lambda x: path.splitext(x)[1])

    OmegaConf.register_new_resolver("scan_name", lambda scan_file: path.join(
        path.dirname(scan_file), f"gaussian_{path.splitext(path.basename(scan_file))[0]}"))

def pretty(cfg):
    return OmegaConf.to_yaml(cfg, resolve=True)